#include <EEPROMex.h>
#include <Wire.h>
#include <Adafruit_GFX.h>
#include <Adafruit_LEDBackpack.h>
#include <FastLED.h>
#include <NewPing.h>

// define the Arduino digital pin connections:
#define RESET_BTN_PIN                    2 // Arduino pin D2 to push button and 10KΩ resistor to ground
#define RESET_LED_PIN                   13 // Arduino pin D13 to on-board LED
#define ECHO_PIN                         3 // Arduino pin D3 to echo pin of HC-SR04 ultrasonic sensor
#define TRIG_PIN                         4 // Arduino pin D4 to trig pin of HC-SR04 ultrasonic sensor
#define RGB_LEDS_PIN                     5 // Arduino pin D5 to RGB strip

// define constants used to control functionality:
#define RGB_LED_COUNT                  116 // number of LEDs in RGB LED strip
#define RGB_LED_TYPE               WS2812B // type of RGB LED strip
#define RGB_LED_COLOR_ORDER            GRB // color order of RGB LED strip
#define RGB_LED_BRIGHTNESS              75 // brightness of RGB LED strip
#define MAX_DISTANCE_CM                 28 // max distance to detect drone (in centimeters)
#define MIN_RACE_TIME_MILLIS          3000 // min race time (in milliseconds)
#define LONG_PRESS_DURATION_MILLIS    1200 // long press duration (in milliseconds)

// define the four different race states:
#define RACE_NOT_STARTED                 0 // race has not started, yet
#define RACE_ACTIVE                      1 // race has started and is active
#define RACE_ENDED_NORMAL                2 // race ended normally
#define RACE_ENDED_RECORD                3 // race ended with a new record time

// define global variables:
Adafruit_7segment fourDigitDisplay = Adafruit_7segment();
CRGB              rgbLeds[RGB_LED_COUNT];
bool              raceStarted;
unsigned long     startMillis;
unsigned long     elapsedMillis;
unsigned long     recordMillis;
unsigned long     droneMillis;
NewPing           distanceSensor(TRIG_PIN, ECHO_PIN, MAX_DISTANCE_CM);

// define constants used to control display of RGB strip:
const CRGB defaultColor = CRGB::Goldenrod;
const CRGB raceColors[] = { CRGB::Green, CRGB::Purple, CRGB::Yellow, CRGB::Blue, CRGB::Orange, CRGB::Pink };
const CRGB finishColor  = CRGB::Red;

// implement setup function that is executed at startup:
void setup() {
  // configure the pin connected to the reset push button:
  pinMode(RESET_BTN_PIN, INPUT);

  // configure the pin connected to the reset LED:
  pinMode(RESET_LED_PIN, OUTPUT);
  digitalWrite(RESET_LED_PIN, LOW);

  // configure the pins connected to the ultrasonic sensor:
  pinMode(ECHO_PIN, INPUT);
  pinMode(TRIG_PIN, OUTPUT);
  digitalWrite(TRIG_PIN, LOW);

  // configure the pin connected to the RGB strip:
  pinMode(RGB_LEDS_PIN, OUTPUT);

  // configure the 4-digit 7-segment display (connected to I2C pins):
  fourDigitDisplay.begin(0x70);

  // configure and update the RGB strip:
  FastLED.addLeds<RGB_LED_TYPE, RGB_LEDS_PIN, RGB_LED_COLOR_ORDER>(rgbLeds, RGB_LED_COUNT);
  FastLED.setBrightness(RGB_LED_BRIGHTNESS);
  updateRgbStrip();

  // reset the race timer:
  resetRaceTimer();

  // set the elapsed time equal to the record time so it displays at startup:
  elapsedMillis = recordMillis;

  // slight delay to avoid 
  delay(100);
}

// implement loop function that is executed repeatedly:
void loop() {
  // if the reset button is down, check for long-press before resetting race timer:
  if(isResetButtonDown()) {
    if(!raceStarted && isResetButtonLongPress()) {
      saveRecordMillis(elapsedMillis);
      blinkLedStrip();
    }
    resetRaceTimer();
  }

  // if the drone is detected within the gate, start or end the race:
  if(isDroneDetected()) {
    if(!raceStarted && (elapsedMillis == 0)) {
      raceStarted = true;
      startMillis = millis();
    }
    else if(raceStarted) {
      raceStarted = false;
    }
  }

  // if the race has started, update the elapsed time:
  if(raceStarted)
    elapsedMillis = millis() - startMillis;

  // display the elapsed time on the four-digit display:
  updateTimerDisplay();

  // display race progress visually on the RGB LED strip:
  updateRgbStrip();
}

void resetRaceTimer() {
  raceStarted   = false;
  startMillis   = 0;
  elapsedMillis = 0;
  droneMillis   = 0;
  recordMillis  = loadRecordMillis();
}

void saveRecordMillis(unsigned long timeInMillis) {
  EEPROM.writeLong(0, timeInMillis);
}

unsigned long loadRecordMillis() {
  return EEPROM.readLong(0);
}

bool isResetButtonDown() {
  bool resetDown = digitalRead(RESET_BTN_PIN);
  digitalWrite(RESET_LED_PIN, resetDown);
  return resetDown;
}

bool isResetButtonLongPress() {
  unsigned long resetMillis = millis();

  while(digitalRead(RESET_BTN_PIN) == HIGH) {
    if((millis() - resetMillis) > LONG_PRESS_DURATION_MILLIS)
      return true;
  }

  return false;
}

void blinkLedStrip() {
  // make a copy of the current LED color array:
  CRGB rgbLedsCopy[RGB_LED_COUNT];
  memcpy(rgbLedsCopy, rgbLeds, sizeof(CRGB)*RGB_LED_COUNT);

  // blink the entire LED strip three times:
  for(int i = 0; i < 3; i++) {
    // turn off all LEDs for a short time:
    fill_solid(rgbLeds, RGB_LED_COUNT, CRGB::Black);
    FastLED.show();
    delay(80);

    // return all LEDs to their original color values for a short time:
    memcpy(rgbLeds, rgbLedsCopy, sizeof(CRGB)*RGB_LED_COUNT);
    FastLED.show();
    delay(80);
  }
}

bool isDroneDetected() {
  // Note: This code ensures a minimum race time has elapsed to avoid stopping the
  //       race if the drone gets hung up in the start gate.

  unsigned long pingTime     = distanceSensor.ping();
  unsigned long distanceInCm = distanceSensor.convert_cm(pingTime);
  if(((pingTime > 0) && (distanceInCm < MAX_DISTANCE_CM)) && (droneMillis == 0)) {
    droneMillis = millis();
    return true;
  }

  if((droneMillis != 0) && (millis() - droneMillis > MIN_RACE_TIME_MILLIS))
    droneMillis = 0;

  return false;
}

int getUltrasonicDistanceInCentimeters() {
  // set the ultrasonic trigger by setting it HIGH for 10 microseconds:
  digitalWrite(TRIG_PIN, HIGH);
  delayMicroseconds(10);
  digitalWrite(TRIG_PIN, LOW);

  // measure the number of microseconds before the echo pin goes HIGH:
  long durationInMicroseconds = pulseIn(ECHO_PIN, HIGH);

  // calculate and return the distance in centimeters based on the speed of sound:
  return durationInMicroseconds * 0.034 / 2;
}

void updateTimerDisplay() {
  // calculate the number of seconds that have elapsed, up to 999.9:
  double raceTime = (double)elapsedMillis / 1000.0;
  if(raceTime > 999.9)
    raceTime = 999.9;

  // update the four digit display to show the race time with one decimal place:
  fourDigitDisplay.printFloat(raceTime, 1 /* fractional digits */, DEC);

  // HACK: add a leading zero and decimal point if race time is less than 1.0:
  if(raceTime < 1.0)
    fourDigitDisplay.writeDigitNum(3, 0, true  /* decimal point */);

  // write the race time to the four-digit display:
  fourDigitDisplay.writeDisplay();
}

void updateRgbStrip() {
  // Note: The RGB strip is folded in half so that color can be seen in all directions.
  //       This approach complicates the code a bit, as LED animations need to be mirrored
  //       so that they look right with an RGB strip that has been folded in half.

  // determine the race mode:
  byte raceMode = RACE_NOT_STARTED;
  if(elapsedMillis > 0) {
    if(raceStarted)
      raceMode = RACE_ACTIVE;
    else if(elapsedMillis >= recordMillis)
      raceMode = RACE_ENDED_NORMAL;
    else
      raceMode = RACE_ENDED_RECORD;
  }

  // color the RGB strip LEDs based on the race mode:
  switch (raceMode) {
    case RACE_NOT_STARTED: {
      // display a solid color while waiting for the race to begin:
      fill_solid(rgbLeds, RGB_LED_COUNT, darkenColor(defaultColor));
      break;
    }

    case RACE_ACTIVE: {
      // Note: During a race, the RGB LED strip changes colors every ten seconds and
      //       shows progress by lighting up LEDs brighter during those ten seconds
      //       from each side.  The code defines 1/4 of the LEDs and mirrors them to
      //       the second 1/4 of the strip.  Finally, the first half of the LED strip
      //       is mirrored to the second half.

      // change active race color every ten seconds:
      int  elapsedSeconds = elapsedMillis / 1000;
      int  totalColors    = sizeof(raceColors) / sizeof(CRGB);
      int  colorIndex     = (elapsedSeconds / 10) % totalColors;
      CRGB raceColor      = raceColors[colorIndex];

      // fill 1/4 of the RGB strip with a darkened race color:
      fill_solid(rgbLeds, RGB_LED_COUNT / 4, darkenColor(raceColor));

      // fill a portion of the RGB strip to indicate progress into ten seconds:
      int  progressCount  = (100 * (elapsedSeconds % 10)) / (RGB_LED_COUNT / 4);
      fill_solid(rgbLeds, progressCount + 1, raceColor);

      // mirror the first 1/4 of the RGB stip to fill 1/2 of the strip:
      mirrorLeds(rgbLeds, RGB_LED_COUNT/2);

      // mirror the first 1/2 of the RGB strip to fill the entire strip:
      mirrorLeds(rgbLeds, RGB_LED_COUNT);
      break;
    }

    case RACE_ENDED_NORMAL: {
      // display a solid color to indicate the race is over (no record):
      fill_solid(rgbLeds, RGB_LED_COUNT, darkenColor(finishColor));
      break;
    }

    case RACE_ENDED_RECORD: {
      // Note: When a race ends that beats the previously saved record time, the
      //       finish color is displayed with a gradient and rotated quickly.

      // create a gradient of the finish color over 10 LEDs:
      CRGB gradientLeds[10];
      fill_gradient_RGB(gradientLeds, 0, finishColor, 9, CRGB::Black);

      // rotate the gradient color array based on the current time:
      CRGB rotatedLeds[10];
      int  rotateAmount = (millis() / 10) % 10;
      for(int i = 0; i < 10; i++)
        rotatedLeds[i] = gradientLeds[(i + rotateAmount) % 10];

      // fill the first half (or more) of the RGB strip with rotated LED colors:
      for(int i = 0; i <= RGB_LED_COUNT/2/10; i++) {
        memcpy(&rgbLeds[i * 10], rotatedLeds, sizeof(CRGB)*10);
      }

      // mirror the first 1/2 of the RGB strip to fill the entire strip:
      mirrorLeds(rgbLeds, RGB_LED_COUNT);
      break;
    }
  }

  // show the RGB LED color array on the RGB strip:
  FastLED.show();
}

CRGB darkenColor(CRGB color) {
  // darken the specified color by dividing each color value by eight:
  return CRGB(color.r/8, color.g/8, color.b/8);
}

void mirrorLeds(CRGB* leds, int count) {
  // mirror the second half of the LEDs to match the first in reverse order:
  for(int i = count/2; i < count; i++) {
    leds[i] = leds[count - i - 1];
  }
}
